\name{rhierLinearModel}
\alias{rhierLinearModel}
\concept{bayes}
\concept{MCMC}
\concept{Gibbs Sampling}
\concept{hierarchical models}
\concept{linear model}

\title{ Gibbs Sampler for Hierarchical Linear Model }
\description{
  \code{rhierLinearModel} implements a Gibbs Sampler for hierarchical linear models with a normal prior.
}
\usage{
rhierLinearModel(Data, Prior, Mcmc, output)
}
\arguments{
  \item{Data}{ list(regdata,Z) (Z optional). }
  \item{Prior}{ list(Deltabar,A,nu.e,ssq,nu,V)  (optional).}
  \item{Mcmc}{ list(R,keep,chains) (R required).}
  \item{output}{ binary output format.} 
}
\details{
  Model: length(regdata) regression equations. \cr
        \eqn{y_i = X_ibeta_i + e_i}. \eqn{e_i} \eqn{\sim}{~} \eqn{N(0,tau_i)}.  nvar X vars in each equation. 

 Priors:\cr
        \eqn{tau_i} \eqn{\sim}{~} nu.e*\eqn{ssq_i/\chi^2_{nu.e}}.  \eqn{tau_i} is the variance of \eqn{e_i}.\cr
        \eqn{beta_i} \eqn{\sim}{~} N(ZDelta[i,],\eqn{V_{beta}}). \cr
               Note:  ZDelta is the matrix Z * Delta; [i,] refers to ith row of this product.

          \eqn{vec(Delta)} given \eqn{V_{beta}} \eqn{\sim}{~} \eqn{N(vec(Deltabar),V_{beta} (x) A^{-1})}.\cr
          \eqn{V_{beta}} \eqn{\sim}{~} \eqn{IW(nu,V)}. \cr
              \eqn{Delta, Deltabar} are nz x nvar.  \eqn{A} is nz x nz.  \eqn{V_{beta}} is nvar x nvar.
        
          Note: The default of Z is iota(nreg x 1).

  List arguments contain:
  \itemize{
    \item{\code{regdata}}{ list of lists with X,y matrices for each of length(regdata) regressions}
    \item{\code{regdata[[i]]$X}}{ X matrix for equation i }
    \item{\code{regdata[[i]]$y}}{ y vector for equation i }
    \item{\code{Deltabar}}{ nz x nvar matrix of prior means (def: 0)}
    \item{\code{A}}{ nz x nz matrix for prior precision (def: .01I)}
    \item{\code{nu.e}}{ d.f. parm for regression error variance prior (def: 3)}
    \item{\code{ssq}}{ scale parm for regression error var prior (def: var(\eqn{y_i}))}
    \item{\code{nu}}{ d.f. parm for Vbeta prior (def: nvar+3)}
    \item{\code{V}}{ Scale location matrix for Vbeta prior (def: nu*I)}
    \item{\code{R}}{ number of MCMC draws}
    \item{\code{keep}}{ MCMC thinning parm: keep every keepth draw (def: 1)}
    \item{\code{chains}}{ number of MCMC simulation chains for \code{CODA} diagnostics}
    \item{\code{output}}{ supported binary output format:
                \code{default}, \code{bayesm}, or \code{coda}.}
   }
}
\value{
For \code{default} format:  a list containing
  \item{betadraw}{R/keep x nreg x nvar array of individual regression coef draws}
  \item{taudraw}{R/keep x nreg array of error variance draws}
  \item{Deltadraw}{R/keep x nz x nvar array of Deltadraws}
  \item{Vbetadraw}{R/keep x nvar*nvar array of Vbeta draws}
  \cr
For \code{bayesm} format:  a list containing
  \item{betadraw}{nreg x nvar x R/keep array of individual regression coef draws}
  \item{taudraw}{R/keep x nreg array of error variance draws}
  \item{Deltadraw}{R/keep x nz x nvar array of Deltadraws}
  \item{Vbetadraw}{R/keep x nvar*nvar array of Vbeta draws}
  \cr
For \code{CODA} output format:
  A list of CODA compatible simulation chains in default format.
  Only the upper triangle of each VbetaDraw sample is kept for coda diagnostics.
}
\references{ 
  Rossi, Allenby and McCulloch:\cr
  \emph{Bayesian Statistics and Marketing} Ch 3\cr
  \url{http://www.perossi.org/home/bsm-1}
}
\author{
  Chi Yau (based on R doc of \code{rhierLinearModel} in \code{bayesm} by Peter Rossi)\cr
  \email{chi.yau@r-tutor.com}
}

\examples{
\dontrun{
library(rpud)
data.path <- file.path(path.package(package="rpud"), "runit/data")

#################################################        
# sample data

Z <- read.table(
        file.path(data.path, "rhierlm-n100-z.txt"), 
        header=FALSE)
Z <- as.matrix(Z)

data <- read.table(
        file.path(data.path, "rhierlm-n100-data.txt"), 
        header=FALSE)
data <- as.matrix(data)

regkey <- unique(data[,1])
colvar <- 3:ncol(data)

regdata <- NULL
for (reg in regkey) {
    
    filter <- (data[,1]==reg)
    y <- data[filter,2]
    X <- data[filter,colvar]
    
    regdata[[reg]] <- list(y=y,X=X)
}

# parameters
R <- 1000
keep <- 1

Data <- list(regdata=regdata,Z=Z)
Mcmc <- list(R=R,keep=keep)

set.seed(66)
out<-rpud::rhierLinearModel(Data=Data,Mcmc=Mcmc,output="bayesm")

library(bayesm)
cat("\n*** Summary of Delta Draws ***",fill=TRUE)
summary(out$Deltadraw)

cat("\n*** Summary of Vbeta Draws ***",fill=TRUE)
summary(out$Vbetadraw)
}}
\keyword{ regression }
